/*
  RPCEmu - An Acorn system emulator

  Copyright (C) 2021 RiscOS Cloverleaf

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Encode/decode UTF8<->RiscOS alphabeths
 * Code mainly taked from NetSurf browser
 * http://www.netsurf-browser.org/
 */

#include <assert.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <iconv.h>
#include "rpcemu.h"

#ifndef NOF_ELEMENTS
#define NOF_ELEMENTS(array) (sizeof(array)/sizeof(*(array)))
#endif


static const char *localencodings[] = {
        "ISO-8859-1//TRANSLIT",	/* BFont - 100 - just use Latin1, instead */
        "ISO-8859-1//TRANSLIT",
        "ISO-8859-2//TRANSLIT",
        "ISO-8859-3//TRANSLIT",
        "ISO-8859-4//TRANSLIT",
        "ISO-8859-5//TRANSLIT",
        "ISO-8859-6//TRANSLIT",
        "ISO-8859-7//TRANSLIT",
        "ISO-8859-8//TRANSLIT",
        "ISO-8859-9//TRANSLIT",
        "ISO-IR-182//TRANSLIT",
        "UTF-8",
        "ISO-8859-15//TRANSLIT",
        "ISO-8859-10//TRANSLIT",
        "ISO-8859-13//TRANSLIT",
        "ISO-8859-14//TRANSLIT",
        "ISO-8859-16//TRANSLIT",
#define CONT_ENC_END 116	/* RISC OS alphabet numbers lie in a
				 * contiguous range [100,CONT_ENC_END]
				 * _except_ for Cyrillic2, which doesn't.
				 */
        "CP866//TRANSLIT"	/* Cyrillic2 - 120 */
};

static const struct special {
    char local;		/**< Local 8bit representation */
    char len;		/**< Length (in bytes) of UTF-8 character */
    const char *utf;	/**< UTF-8 representation */
} special_chars[] = {
        { 0x80, 3, "\xE2\x82\xAC" },	/* EURO SIGN */
        { 0x81, 2, "\xC5\xB4" },	/* LATIN CAPITAL LETTER W WITH CIRCUMFLEX */
        { 0x82, 2, "\xC5\xB5" },	/* LATIN SMALL LETTER W WITH CIRCUMFLEX */
        { 0x84, 3, "\xE2\x9C\x98" },	/* HEAVY BALLOT X */
        { 0x85, 2, "\xC5\xB6" },	/* LATIN CAPITAL LETTER Y WITH CIRCUMFLEX */
        { 0x86, 2, "\xC5\xB7" },	/* LATIN SMALL LETTER Y WITH CIRCUMFLEX */
        { 0x88, 3, "\xE2\x87\x90" },	/* LEFTWARDS DOUBLE ARROW */
        { 0x89, 3, "\xE2\x87\x92" },	/* RIGHTWARDS DOUBLE ARROW */
        { 0x8a, 3, "\xE2\x87\x93" },	/* DOWNWARDS DOUBLE ARROW */
        { 0x8b, 3, "\xE2\x87\x91" },	/* UPWARDS DOUBLE ARROW */
        { 0x8c, 3, "\xE2\x80\xA6" },	/* HORIZONTAL ELLIPSIS */
        { 0x8d, 3, "\xE2\x84\xA2" },	/* TRADE MARK SIGN */
        { 0x8e, 3, "\xE2\x80\xB0" },	/* PER MILLE SIGN */
        { 0x8f, 3, "\xE2\x80\xA2" },	/* BULLET */
        { 0x90, 3, "\xE2\x80\x98" },	/* LEFT SINGLE QUOTATION MARK */
        { 0x91, 3, "\xE2\x80\x99" },	/* RIGHT SINGLE QUOTATION MARK */
        { 0x92, 3, "\xE2\x80\xB9" },	/* SINGLE LEFT-POINTING ANGLE QUOTATION MARK */
        { 0x93, 3, "\xE2\x80\xBA" },	/* SINGLE RIGHT-POINTING ANGLE QUOTATION MARK */
        { 0x94, 3, "\xE2\x80\x9C" },	/* LEFT DOUBLE QUOTATION MARK */
        { 0x95, 3, "\xE2\x80\x9D" },	/* RIGHT DOUBLE QUOTATION MARK */
        { 0x96, 3, "\xE2\x80\x9E" },	/* DOUBLE LOW-9 QUOTATION MARK */
        { 0x97, 3, "\xE2\x80\x93" },	/* EN DASH */
        { 0x98, 3, "\xE2\x80\x94" },	/* EM DASH */
        { 0x99, 3, "\xE2\x88\x92" },	/* MINUS SIGN */
        { 0x9a, 2, "\xC5\x92" },	/* LATIN CAPITAL LIGATURE OE */
        { 0x9b, 2, "\xC5\x93" },	/* LATIN SMALL LIGATURE OE */
        { 0x9c, 3, "\xE2\x80\xA0" },	/* DAGGER */
        { 0x9d, 3, "\xE2\x80\xA1" },	/* DOUBLE DAGGER */
        { 0x9e, 3, "\xEF\xAC\x81" },	/* LATIN SMALL LIGATURE FI */
        { 0x9f, 3, "\xEF\xAC\x82" } 	/* LATIN SMALL LIGATURE FL */
};

// typedef unsigned long uint32_t;

/* Cache of previous iconv conversion descriptor used by utf8_convert */
static struct {
    char from[32];	/**< Encoding name to convert from */
    char to[32];	/**< Encoding name to convert to */
    iconv_t cd;	/**< Iconv conversion descriptor */
} last_cd;

static inline void utf8_clear_cd_cache(void)
{
    last_cd.from[0] = '\0';
    last_cd.to[0] = '\0';
    last_cd.cd = 0;
}

/**
 * Convert a string from one encoding to another
 *
 * \param in_str  The NULL-terminated string to convert
 * \param len     Length of input string to consider (in bytes), or 0
 * \param from    The encoding name to convert from
 * \param to      The encoding name to convert to
 * \param result  Pointer to location in which to store result.
 * \param result_len Pointer to location in which to store result length.
 * \return NSERROR_OK for no error, NSERROR_NOMEM on allocation error,
 *         NSERROR_BAD_ENCODING for a bad character encoding
 */
static void
utf8_convert(const char *in_str,
             size_t len,
             const char *from,
             const char *to,
             char **result,
             size_t *result_len)
{
    iconv_t cd;
    char *temp, *out, *in, *error_str;
    size_t slen, rlen;

    assert(in_str && from && to && result);

    if (in_str[0] == '\0') {
        /* On AmigaOS, iconv() returns an error if we pass an
         * empty in_str.  This prevents iconv() being called as
         * there is no conversion necessary anyway. */
        *result = strdup("");
        if (!(*result)) {
            error_str = "strdup('') failed";
            goto fail;
        }
        return;
    }

    if (strcasecmp(from, to) == 0) {
        /* conversion from an encoding to itself == strdup */
        *(result) = strdup(in_str);
        if (!(*result)) {
            error_str = "strdup(in_str) failed";
            goto fail;
        }
        return;
    }

    in = (char *)in_str;

    /* we cache the last used conversion descriptor,
     * so check if we're trying to use it here */
    if (strncasecmp(last_cd.from, from, sizeof(last_cd.from)) == 0 &&
        strncasecmp(last_cd.to, to, sizeof(last_cd.to)) == 0) {
        cd = last_cd.cd;
    }
    else {
        /* no match, so create a new cd */
        cd = iconv_open(to, from);
        if (cd == (iconv_t)-1) {
            if (errno == 22) {
                error_str = "bad_encoding";
                goto fail;
            }
            error_str = "iconv_open failed";
            goto fail;
        }

        /* close the last cd - we don't care if this fails */
        if (last_cd.cd)
            iconv_close(last_cd.cd);

        /* and copy the to/from/cd data into last_cd */
        snprintf(last_cd.from, sizeof(last_cd.from), "%s", from);
        snprintf(last_cd.to, sizeof(last_cd.to), "%s", to);
        last_cd.cd = cd;
    }

    slen = len ? len : strlen(in_str);
    /* Worst case = ASCII -> UCS4, so allocate an output buffer
     * 4 times larger than the input buffer, and add 4 bytes at
     * the end for the NULL terminator
     */
    rlen = slen * 4 + 4;

    temp = out = (char*) malloc(rlen);
    if (!out) {
        error_str = "malloc(rlen) failed";
        goto fail;
    }

    /* perform conversion */
    if (iconv(cd, &in, &slen, &out, &rlen) == (size_t)-1) {
        free(temp);
        /* clear the cached conversion descriptor as it's invalid */
        if (last_cd.cd)
            iconv_close(last_cd.cd);
        utf8_clear_cd_cache();
        /** \todo handle the various cases properly
         * There are 3 possible error cases:
         * a) Insufficiently large output buffer
         * b) Invalid input byte sequence
         * c) Incomplete input sequence */
        error_str = "iconv() failed";
        goto fail;
    }

    *(result) = (char*)realloc(temp, out - temp + 4);
    if (!(*result)) {
        free(temp);
        error_str = "realloc() failed";
        goto fail;
    }

    /* NULL terminate - needs 4 characters as we may have
     * converted to UTF-32 */
    memset((*result) + (out - temp), 0, 4);

    if (result_len != NULL) {
        *result_len = (out - temp);
    }
    return;

fail:
    *result = NULL;
    rpclog("utf8_convert: '%s' %s->%s: %s\n", in_str, from, to, error_str);
#if defined __linux || defined __linux__
    fprintf(stderr, "utf8_convert: '%s' %s->%s: %s\n", in_str, from, to, error_str);
#endif
    return;
}

void utf8_to_enc(const char *in_str, const char *encname,
                 size_t len, char **result)
{
    utf8_convert(in_str, len, "UTF-8", encname, result, NULL);
}

void utf8_from_enc(const char *in_str, const char *encname,
                   size_t len, char **result, size_t *result_len)
{
    utf8_convert(in_str, len, encname, "UTF-8", result, result_len);
}

/**
 * Find next legal UTF-8 char in string
 *
 * \param s        The string (assumed valid)
 * \param len      Maximum offset in string
 * \param off      Offset in the string to start at
 * \param nextoff  Pointer to location to receive offset of first byte of
 *                 next legal character
 * \param error    Location to receive error code
 */
#define UTF8_NEXT(s, len, off, nextoff, error)                          \
do {                                                                    \
        if (s == NULL || off >= len) {               \
                error = 1;                            \
                break;                                                  \
        }                                                               \
                                                                        \
        /* Skip current start byte (if present - may be mid-sequence) */\
        if (s[off] < 0x80 || (s[off] & 0xC0) == 0xC0)                   \
                off++;                                                  \
                                                                        \
        while (off < len && (s[off] & 0xC0) == 0x80)                    \
                off++;                                                  \
                                                                        \
        *nextoff = off;                                                 \
                                                                        \
        error = 0;                                         \
} while(0)

size_t utf8_next(const unsigned char s[], size_t l, size_t o)
{
    uint32_t next = 0;
    int perror;

    UTF8_NEXT(s, l, o, &next, perror);
    assert(perror == 0);

    return next;
}


/**
 * Convert a UTF-8 encoded string into the system local encoding
 *
 * \param string The string to convert
 * \param len The length (in bytes) of the string, or 0
 * \param result Pointer to location in which to store result
 */
void utf8_to_local_encoding(const char *in_str, int alphabet, char **result)
{
    int i;
    size_t off, prev_off, len;
    char *temp, *cur_pos, *error_str;
    const char *enc;

    assert(in_str);
    assert(result);

    /* UTF-8 -> simply copy in_str */
    if (alphabet == 111 /* UTF-8 */) {
        *result = strdup(in_str);
        return;
    }

    /* get encoding name */
    enc = (alphabet <= CONT_ENC_END ? localencodings[alphabet - 100]
                                    : (alphabet == 120 ?
                                       localencodings[CONT_ENC_END - 100 + 1]
                                                       : localencodings[0]));

    /* create output buffer */
    len = strlen(in_str);
    *(result) = (char*) malloc(len + 1);
    if (!(*result)) {
        error_str = "malloc() failed";
        goto fail;
    }
    *(*result) = '\0';

    prev_off = 0;
    cur_pos = (*result);

    /* Iterate over string, converting input between unconvertable
     * characters and inserting appropriate output for characters
     * that iconv can't handle. */
    for (off = 0; off < len; off = utf8_next((const unsigned char*)in_str, len, off)) {
        if ((unsigned char)in_str[off] != 0xE2 &&
                (unsigned char)in_str[off] != 0xC5 && (unsigned char)in_str[off] != 0xEF)
            continue;

        for (i = 0; i != NOF_ELEMENTS(special_chars); i++) {
            if (strncmp(in_str + off, special_chars[i].utf,
                        special_chars[i].len) != 0)
                continue;

            /* 0 length has a special meaning to utf8_to_enc */
            if (off - prev_off > 0) {
                utf8_to_enc(in_str + prev_off, enc,
                            off - prev_off, &temp);
                if (!(*temp)) {
                    free(*result);
                    error_str = "utf8_to_enc() failed";
                    goto fail;
                }

                strcat(cur_pos, temp);
                cur_pos += strlen(temp);
                free(temp);
            }

            *cur_pos = special_chars[i].local;
            *(++cur_pos) = '\0';
            prev_off = off + special_chars[i].len;
        }
    }

    /* handle last chunk
     * NB. 0 length has a special meaning to utf8_to_enc */

    if (prev_off < len) {
        utf8_to_enc(in_str + prev_off, enc, len - prev_off,
                    &temp);
        if (!*temp) {
            free(*result);
            error_str = "utf8_to_enc() failed";
            goto fail;
        }

        strcat(cur_pos, temp);

        free(temp);
    }
    return;
fail:
    *result = NULL;
    rpclog("utf8_to_local_encoding: '%s': %s\n", in_str, error_str);
#if defined __linux || defined __linux__
    fprintf(stderr, "utf8_to_local_encoding: '%s': %s\n", in_str, error_str);
#endif
    return;
}

/**
 * Convert a string encoded in the system local encoding to UTF-8
 *
 * \param in_str The string to convert
 * \param len The length (in bytes) of the string, or 0
 * \param result Pointer to location in which to store result
 */
void utf8_from_local_encoding(const char *in_str, int alphabet, char **result)
{
    int i, num_specials = 0, result_alloc;
#define SPECIAL_CHUNK_SIZE 255
    size_t off, prev_off, cur_off, len;
    char *temp, *error_str;
    const char *enc;

    assert(in_str && result);

    /* UTF-8 -> simply copy in_str */
    if (alphabet == 111 /* UTF-8 */) {
        temp = strdup(in_str);
        if (!temp) {
            error_str = "strdup() failed";
            goto fail;
        }
        *result = temp;
        return;
    }

    /* get encoding name */
    enc = (alphabet <= CONT_ENC_END ? localencodings[alphabet - 100]
                                    : (alphabet == 120 ?
                                       localencodings[CONT_ENC_END - 100 + 1]
                                                       : localencodings[0]));

    len = strlen(in_str);
    /* create output buffer (oversized) */
    result_alloc = (len * 4) + (3 * SPECIAL_CHUNK_SIZE) + 1;

    *(result) = (char*)malloc(result_alloc);
    if (!(*result)) {
        error_str = "malloc() failed";
        goto fail;
    }

    *(*result) = '\0';

    prev_off = 0;
    cur_off = 0;

    /* Iterate over string, converting input between unconvertable
     * characters and inserting appropriate output for characters
     * that iconv can't handle. */
    for (off = 0; off < len; off++) {
        if ((unsigned char)in_str[off] < 0x80 || (unsigned char)in_str[off] > 0x9f)
            continue;

        for (i = 0; i != NOF_ELEMENTS(special_chars); i++) {
            if (in_str[off] != special_chars[i].local)
                continue;

            /* 0 length has a special meaning to utf8_from_enc */
            if (off - prev_off > 0) {
                utf8_from_enc(in_str + prev_off, enc,
                              off - prev_off, &temp, NULL);
                if (!*temp) {
                    free(*result);
                    error_str = "utf8_from_enc() failed";
                    goto fail;
                }

                strcat((*result) + cur_off, temp);
                cur_off += strlen(temp);
                free(temp);
            }

            strcat((*result) + cur_off, special_chars[i].utf);

            cur_off += special_chars[i].len;

            prev_off = off + 1;

            num_specials++;
            if (num_specials % SPECIAL_CHUNK_SIZE ==
                SPECIAL_CHUNK_SIZE - 1) {
                char *temp = (char*)realloc((*result),
                                            result_alloc +
                                            (3 * SPECIAL_CHUNK_SIZE));
                if (!temp) {
                    free(*result);
                    error_str = "realloc() failed";
                    goto fail;
                }

                *result = temp;
                result_alloc += (3 * SPECIAL_CHUNK_SIZE);
            }
        }
    }

    /* handle last chunk
     * NB. 0 length has a special meaning to utf8_from_enc */
    if (prev_off < len) {
        utf8_from_enc(in_str + prev_off, enc, len - prev_off,
                          &temp, NULL);
        if (!*temp) {
            free(*result);
            error_str = "utf8_from_enc() failed";
            goto fail;
        }

        strcat((*result) + cur_off, temp);

        cur_off += strlen(temp);

        free(temp);
    }

    /* and copy into more reasonably-sized buffer */
    temp = (char*)realloc((*result), cur_off + 1);
    if (!temp) {
        free(*result);
        error_str = "realloc() failed";
        goto fail;
    }
    *result = temp;
    return;

fail:
    *result = NULL;
    rpclog("utf8_from_local_encoding: '%s': %s\n", in_str, error_str);
#if defined __linux || defined __linux__
    fprintf(stderr, "utf8_from_local_encoding: '%s': %s\n", in_str, error_str);
#endif
    return;
}